@@ -29,6 +29,8 @@
 #include <curl/curl.h>
 #include <curl/mprintf.h>
 
+#include <locale.h> /* for setlocale() */
+
 #include "version.h"
 
 #ifdef _MSC_VER
@@ -48,6 +50,9 @@
 #else
 #define CURLU_ALLOW_SPACE 0
 #endif
+#if CURL_AT_LEAST_VERSION(7,88,0)
+#define SUPPORTS_PUNYCODE
+#endif
 
 #define OUTPUT_URL      0  /* default */
 #define OUTPUT_SCHEME   1
@@ -134,6 +139,16 @@ static void errorf(int exit_code, char *fmt, ...)
   exit(exit_code);
 }
 
+#define VERIFY(op, exit_code, ...) \
+  do { \
+    if(op->verify) { \
+      /* make sure to terminate the JSON array */ \
+      if(op->jsonout) \
+        fputs("\n]\n", stdout); \
+      errorf(exit_code, __VA_ARGS__); \
+    } else \
+      warnf(__VA_ARGS__); \
+  } while(0)
 
 static void help(void)
 {
@@ -167,8 +182,8 @@ static void help(void)
 static void show_version(void)
 {
   curl_version_info_data *data = curl_version_info(CURLVERSION_NOW);
-  fprintf(stdout, "%s version %s libcurl/%s\n", PROGNAME, TRURL_VERSION_TXT,
-          data->version);
+  fprintf(stdout, "%s version %s libcurl/%s [built-with %s]\n",
+          PROGNAME, TRURL_VERSION_TXT, data->version, LIBCURL_VERSION);
   exit(0);
 }
 
@@ -208,6 +223,12 @@ char *qpairs[MAX_QPAIRS]; /* encoded */
 char *qpairsdec[MAX_QPAIRS]; /* decoded */
 int nqpairs; /* how many is stored */
 
+static char *strurldecode(const char *url, int inlength, int *outlength)
+{
+  return curl_easy_unescape(NULL, inlength ? url : "", inlength,
+                            outlength);
+}
+
 static void urladd(struct option *o, const char *url)
 {
   struct curl_slist *n;
@@ -464,6 +485,8 @@ static void get(struct option *op, CURLU *uh)
         char *cl;
         size_t vlen;
         bool urldecode = true;
+        bool punycode = false;
+        bool handled = true;
         end = strchr(ptr, endbyte);
         ptr++; /* pass the { */
         if(!end) {
@@ -485,10 +508,21 @@ static void get(struct option *op, CURLU *uh)
             showqkey(&ptr[10], end - cl - 1, urldecode, true);
           else if(!strncmp(ptr, "query:", 6))
             showqkey(&ptr[6], end - cl - 1, urldecode, false);
+          else if(!strncmp(ptr, "puny:", 5)) {
+            punycode = true;
+#ifndef SUPPORTS_PUNYCODE
+            warnf("Built without punycode support");
+#endif
+            ptr = cl + 1;
+            vlen = end - ptr;
+            handled = false;
+          }
           else
             errorf(ERROR_GET, "Bad --get syntax: %s", ptr);
         }
-        else {
+        else
+          handled = false;
+        if(!handled) {
           const struct var *v = comp2var(ptr, vlen);
           if(v) {
             char *nurl;
@@ -496,6 +530,9 @@ static void get(struct option *op, CURLU *uh)
             rc = curl_url_get(uh, v->part, &nurl,
                               CURLU_DEFAULT_PORT|
                               CURLU_NO_DEFAULT_PORT|
+#ifdef SUPPORTS_PUNYCODE
+                              (punycode?CURLU_PUNYCODE:0)|
+#endif
                               (urldecode?CURLU_URLDECODE:0));
             switch(rc) {
             case CURLUE_OK:
@@ -560,12 +597,10 @@ static void get(struct option *op, CURLU *uh)
   fputc('\n', stream);
 }
 
-static const struct var *setone(CURLU *uh, struct option *o,
-                                const char *setline)
+static const struct var *setone(CURLU *uh, const char *setline)
 {
   char *ptr = strchr(setline, '=');
   const struct var *v = NULL;
-  (void)o;
   if(ptr && (ptr > setline)) {
     size_t vlen = ptr - setline;
     bool urlencode = true;
@@ -600,7 +635,7 @@ static void set(CURLU *uh,
   for(node =  o->set_list; node; node = node->next) {
     const struct var *v;
     char *set = node->data;
-    v = setone(uh, o, set);
+    v = setone(uh, set);
     if(v) {
       if(mask & (1 << v->part))
         errorf(ERROR_SET, "duplicate --set for component %s", v->name);
@@ -613,12 +648,7 @@ static void jsonString(FILE *stream, const char *in, size_t len,
                        bool lowercase)
 {
   const unsigned char *i = (unsigned char *)in;
-  const char *in_end;
-
-  if(!len)
-    in_end = in + strlen(in);
-  else
-    in_end = &in[len];
+  const char *in_end = &in[len];
 
   fputc('\"', stream);
   for(; i < (unsigned char *)in_end; i++) {
@@ -663,6 +693,7 @@ static void jsonString(FILE *stream, const char *in, size_t len,
 static void json(struct option *o, CURLU *uh)
 {
   int i;
+  bool first = true;
   (void)o;
   printf("%s  {\n", o->urls?",\n":"");
   for(i = 0; variables[i].name; i++) {
@@ -671,26 +702,34 @@ static void json(struct option *o, CURLU *uh)
                                 (i?CURLU_DEFAULT_PORT:0)|
                                 CURLU_URLDECODE);
     if(!rc) {
-      if(i)
+      if(!first)
         fputs(",\n", stdout);
+      first = false;
       printf("    \"%s\": ", variables[i].name);
-      jsonString(stdout, nurl, 0, false);
+      jsonString(stdout, nurl, strlen(nurl), false);
       curl_free(nurl);
     }
   }
+  first = true;
   if(nqpairs) {
     int i;
     fputs(",\n    \"params\": [\n", stdout);
     for(i = 0 ; i < nqpairs; i++) {
-      char *sep = strchr(qpairsdec[i], '=');
-      if(i)
+      const char *sep = strchr(qpairsdec[i], '=');
+      const char *value = sep ? sep + 1 : "";
+
+      /* don't print out empty/trimmed values */
+      if(!qpairsdec[i][0])
+          continue;
+      if(!first)
         fputs(",\n", stdout);
+      first = false;
       fputs("      {\n        \"key\": ", stdout);
       jsonString(stdout, qpairsdec[i],
                  sep ? (size_t)(sep - qpairsdec[i]) : strlen(qpairsdec[i]),
                  false);
       fputs(",\n        \"value\": ", stdout);
-      jsonString(stdout, sep ? sep + 1 : "", 0, false);
+      jsonString(stdout, value, strlen(value), false);
       fputs("\n      }", stdout);
     }
     fputs("\n    ]", stdout);
@@ -767,13 +806,12 @@ static char *memdupdec(char *source, size_t len)
   int leftlen = 0;
   int rightlen = 0;
 
-  left = curl_easy_unescape(NULL, source, sep ? (size_t)(sep - source) : len,
-                            &leftlen);
+  left = strurldecode(source, sep ? (size_t)(sep - source) : len,
+                      &leftlen);
   if(sep) {
     char *p;
     int plen;
-    right = curl_easy_unescape(NULL, sep + 1, len - (sep - source) - 1,
-                               &rightlen);
+    right = strurldecode(sep + 1, len - (sep - source) - 1, &rightlen);
 
     /* convert null bytes to periods */
     for(plen = rightlen, p = right; plen; plen--, p++) {
@@ -887,6 +925,16 @@ static void sortquery(struct option *o)
   }
 }
 
+static CURLUcode urlfromstring(struct option *o,
+                               CURLU *uh,
+                               const char *url)
+{
+  return curl_url_set(uh, CURLUPART_URL, url,
+                      CURLU_GUESS_SCHEME|CURLU_NON_SUPPORT_SCHEME|
+                      (o->accept_space ?
+                       (CURLU_ALLOW_SPACE|CURLU_URLENCODE) : 0));
+}
+
 static void singleurl(struct option *o,
                       const char *url, /* might be NULL */
                       struct iterinfo *iinfo,
@@ -898,15 +946,9 @@ static void singleurl(struct option *o,
     if(!uh)
       errorf(ERROR_MEM, "out of memory");
     if(url) {
-      CURLUcode rc =
-        curl_url_set(uh, CURLUPART_URL, url,
-                     CURLU_GUESS_SCHEME|CURLU_NON_SUPPORT_SCHEME|
-                     (o->accept_space ?
-                      (CURLU_ALLOW_SPACE|CURLU_URLENCODE) : 0));
+      CURLUcode rc = urlfromstring(o, uh, url);
       if(rc) {
-        if(o->verify)
-          errorf(ERROR_BADURL, "%s [%s]", curl_url_strerror(rc), url);
-        warnf("%s [%s]", curl_url_strerror(rc), url);
+        VERIFY(o, ERROR_BADURL, "%s [%s]", curl_url_strerror(rc), url);
         return;
       }
       else {
@@ -919,6 +961,7 @@ static void singleurl(struct option *o,
   do {
     char iterbuf[1024];
     struct curl_slist *p;
+    bool url_is_invalid = false;
     /* set everything */
     set(uh, o);
 
@@ -972,7 +1015,7 @@ static void singleurl(struct option *o,
       curl_msnprintf(iterbuf, sizeof(iterbuf), "%.*s%s=%.*s", (int)plen, part,
                      urlencode ? "" : ":",
                      (int)wlen, w);
-      setone(uh, o, iterbuf);
+      setone(uh, iterbuf);
       if(iter && iter->next) {
         struct iterinfo info;
         memset(&info, 0, sizeof(info));
@@ -1022,8 +1065,39 @@ static void singleurl(struct option *o,
     /* put the query back */
     qpair2query(uh, o);
 
+    /* make sure the URL is still valid */
+    if(!url || o->redirect || o->set_list || o->append_path) {
+      char *ourl = NULL;
+      CURLUcode rc = curl_url_get(uh, CURLUPART_URL, &ourl, 0);
+      if(rc) {
+        VERIFY(o, ERROR_URL, "not enough input for a URL");
+        url_is_invalid = true;
+      }
+      else {
+        rc = urlfromstring(o, uh, ourl);
+        if(rc) {
+          VERIFY(o, ERROR_BADURL, "%s [%s]", curl_url_strerror(rc),
+                 ourl);
+          url_is_invalid = true;
+        }
+        else {
+          char *nurl = NULL;
+          rc = curl_url_get(uh, CURLUPART_URL, &nurl, 0);
+          if(rc || strcmp(ourl, nurl)) {
+            VERIFY(o, ERROR_BADURL, "url became invalid");
+            url_is_invalid = true;
+          }
+          if(!rc)
+            curl_free(nurl);
+        }
+        curl_free(ourl);
+      }
+    }
+
     if(iter && iter->next)
       ;
+    else if(url_is_invalid)
+      ;
     else if(o->jsonout)
       json(o, uh);
     else if(o->format) {
@@ -1037,9 +1111,6 @@ static void singleurl(struct option *o,
         printf("%s\n", nurl);
         curl_free(nurl);
       }
-      else {
-        errorf(ERROR_URL, "not enough input for a URL");
-      }
     }
 
     fflush(stdout);
@@ -1059,6 +1130,7 @@ int main(int argc, const char **argv)
   struct option o;
   struct curl_slist *node;
   memset(&o, 0, sizeof(o));
+  setlocale(LC_ALL, "");
   curl_global_init(CURL_GLOBAL_ALL);
 
   for(argc--, argv++; argc > 0; argc--, argv++) {
@@ -1133,6 +1205,7 @@ int main(int argc, const char **argv)
       else {
         struct iterinfo iinfo;
         memset(&iinfo, 0, sizeof(iinfo));
+        o.verify = true;
         singleurl(&o, NULL, &iinfo, o.iter_list);
       }
     } while(node);
